module DictionarySlicing

using OrderedCollections

export collectall, slice, directslicing

"""
    collectall(sets...; maxdepth=5)

Recursively collects elements from `sets`.

Keyword args:
maxdepth --------- determines maximum recursion level
currentdepth ----- currect recursion level
"""
function collectall(args...; maxdepth = 5, currentdepth = 1)	# if you are reading this, where should this go? I feel like it belongs elsewhere...
	ff = []
	f1 = collect([args...])
	f2 = collect([Base.Iterators.flatten(f1)...])
	for x in f2
		if typeof(x) <: Union{AbstractArray,AbstractRange} && currentdepth < maxdepth
			xx = collect([Base.Iterators.flatten(x)...])
			push!(ff,collectall(xx; maxdepth, currentdepth = currentdepth+1)...)
		else
			push!(ff,x)
		end
	end
	return ff
end

"""
    slice(D::T, idxs...; keep = :first, filter = nothing) where {T<:AbstractDict}

Returns the slices from dict `D`.

Keyword args:
keep --------- either keep the :first or :last instance of a slice
filter ------- true/false-returning function to apply to collected indices
"""
function slice(D::T, idxs...; keep = :first, filter = nothing) where {T<:AbstractDict}
	indices = collectall(idxs...)

	if filter != nothing
		try
			if keep == :lastbefore || keep == "lastbefore"
				revindices = reverse(indices)
				revindices = Base.filter(filter,revindices)
				dd = OrderedDict([([keys(D)...][i],[values(D)...][i]) for i in revindices])
				rv = length(dd):-1:1
				return slice(dd,rv...)
			elseif keep == :firstbefore || keep == "firstbefore"
				indices = unique(indices)
				return slice(D,indices...; filter = filter)
			else
				indices = Base.filter(filter,indices)
			end
		catch
			error("could not filter")
		end
	end
	if keep == :last || keep == "last"
		revindices = reverse(indices)
		dd = OrderedDict([([keys(D)...][i],[values(D)...][i]) for i in revindices])
		rv = length(dd):-1:1
		return slice(dd,rv...)
	end
	ks = [[keys(D)...][i] for i in indices]
	vs = [[values(D)...][i] for i in indices]

    return T(ks.=>vs)
end
"""
    directslicing()

Activates indexing/slicing for OrderedDict with function indexing.
"""
function directslicing()
	@eval(function (D::OrderedDict)(idxs...; keep = :first, filter = nothing)
		indices = collectall(idxs...)

		if filter != nothing
			try
				if keep == :lastbefore || keep == "lastbefore"
					revindices = reverse(indices)
					revindices = Base.filter(filter,revindices)
					dd = OrderedDict([([keys(D)...][i],[values(D)...][i]) for i in revindices])
					rv = length(dd):-1:1
					return dd(rv...)
				elseif keep == :firstbefore || keep == "firstbefore"
					indices = unique(indices)
					return D(indices...; filter = filter)
				else
					indices = Base.filter(filter,indices)
				end
			catch
				error("could not filter")
			end
		end
		if keep == :last || keep == "last"
			revindices = reverse(indices)
			dd = OrderedDict([([keys(D)...][i],[values(D)...][i]) for i in revindices])
			rv = length(dd):-1:1
			return dd(rv...)
		end

	    return OrderedDict([([keys(D)...][i],[values(D)...][i]) for i in indices])
	end)
end

end
